// Flutter Packages
import 'package:flutter/material.dart';

// Dart Packages
import 'dart:async';

// Firebase Packages
import 'package:firebase_database/firebase_database.dart';

// Data Models
import 'models/view_type.dart';
import 'models/wrap_options.dart';

// Widgets
import 'widgets/build_pagination.dart';

// Functions
import 'functions/fetch_data.dart';

class RealtimeDBPagination extends StatefulWidget {
  const RealtimeDBPagination({
    Key? key,
    required this.query,
    required this.itemBuilder,
    this.startAfter,
    this.limit = 10,
    this.viewType = ViewType.list,
    this.isRealtime = true,
    this.gridDelegate = const SliverGridDelegateWithFixedCrossAxisCount(
      crossAxisCount: 2,
      childAspectRatio: 1,
    ),
    this.wrapOptions = const WrapOptions(),
  }) : super(key: key);

  final Query query;
  final Widget Function(BuildContext, DataSnapshot, int) itemBuilder;
  final DataSnapshot? startAfter;
  final int limit;
  final ViewType viewType;
  final bool isRealtime;
  final SliverGridDelegate gridDelegate;
  final WrapOptions wrapOptions;

  @override
  State<RealtimeDBPagination> createState() => _RealtimeDBPaginationState();
}

class _RealtimeDBPaginationState extends State<RealtimeDBPagination> {
  List<DataSnapshot> _data = [];
  StreamSubscription<DatabaseEvent>? _streamSub;
  ScrollController controller = ScrollController();
  bool isLoading = true;
  bool isFetching = false;

  Future<void> _loadData() async {
    setState(() {
      isFetching = true;
    });

    final List<DataSnapshot> documents = await fetchData(
      query: widget.query,
      limit: widget.limit,
      startAfter: widget.startAfter ?? _data.last,
    );

    _data += documents;

    isFetching = false;
    isLoading = false;

    if (mounted) {
      setState(() {});
    }

    _setAutoUpdateListener();
  }

  void _setAutoUpdateListener() async {
    await _streamSub?.cancel();
    _streamSub = widget.query
        .limitToFirst(_data.length + (widget.isRealtime ? 1 : 0))
        .onValue
        .listen(
      (DatabaseEvent snapshot) {
        _data = snapshot.snapshot.children.map((DataSnapshot document) {
          return document;
        }).toList();

        if (mounted) {
          setState(() {});
        }
        _setAutoUpdateListener();
      },
    );
  }

  void _scrollListener() {
    final ScrollPosition position = controller.position;
    if (position.pixels >= position.maxScrollExtent &&
        !isLoading &&
        !isFetching) {
      _loadData();
    }
  }

  @override
  void initState() {
    super.initState();
    _loadData();

    controller.addListener(_scrollListener);
  }

  @override
  void dispose() {
    _streamSub?.cancel();
    controller.removeListener(_scrollListener);
    controller.dispose();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return isLoading
        ? const Center(child: CircularProgressIndicator.adaptive())
        : BuildPagination(
            controller: controller,
            items: _data,
            itemBuilder: widget.itemBuilder as dynamic,
            isLoading: isLoading,
          );
  }
}
